"""


## Add head configuration

This snippet allows users to configure their head in `markata.toml`.

``` html
{{ config.get('head', {}).pop('text') if 'text' in config.get('head',{}).keys() }}{% for tag, meta in config.get('head', {}).items() %}{% for _meta in meta %}
<{{ tag }} {% for attr, value in _meta.items() %}{{ attr }}="{{ value }}"{% endfor %}/> {% endfor %}{% endfor %}
```

Users can specify any sort of tag in their `markata.toml`

``` toml
[[markata.head.meta]]
name = "og:type"
content = "article"

[[markata.head.meta]]
name = "og:author"
content = "Waylon Walker"
```

The above configuration becomes this once rendered.

``` html
<meta name='og:type' content='article' />
<meta name='og:Author' content='Waylon Walker' />
```

!! Note

    Article variables can be used for dynamic entries like canonical_url
    ``` toml
    [markata]
    url = "markata.dev"

    [[markata.head.meta]]
    href="{{ config.url }}/{{ slug }}/"
    rel="canonical"
    ```

Optionally users can also specify plain text to be appended to the head of
their documents.  This works well for things that involve full blocks.

``` toml
[[markata.head.text]]
value = '''
<script>
    console.log('hello world')
</script>
'''

[[markata.head.text]]
value='''
html  {
    font-family: "Space Mono", monospace;
    background: var(--color-bg);
    color: var(--color-text);
}
'''
```

"""
import copy
from pathlib import Path
from typing import TYPE_CHECKING

import jinja2
from deepmerge import always_merger
from jinja2 import Template, Undefined
from more_itertools import flatten

from markata import __version__
from markata.hookspec import hook_impl

env = jinja2.Environment()

if TYPE_CHECKING:
    from markata import Markata


class SilentUndefined(Undefined):
    def _fail_with_undefined_error(self, *args, **kwargs):
        return ""


@hook_impl
def configure(markata: "Markata") -> None:
    """
    Massages the configuration limitations of toml to allow a little bit easier
    experience to the end user making configurations while allowing an simpler
    jinja template.  This enablees the use of the `markata.head.text` list in
    configuration.
    """

    raw_text = markata.config.get("head", {}).get("text", "")

    if isinstance(raw_text, list):
        markata.config["head"]["text"] = "\n".join(
            flatten([t.values() for t in raw_text])
        )


@hook_impl
def pre_render(markata: "Markata") -> None:
    """
    FOR EACH POST: Massages the configuration limitations of toml/yaml to allow
    a little bit easier experience to the end user making configurations while
    allowing an simpler jinja template.  This enablees the use of the
    `markata.head.text` list in configuration.
    """
    for article in [a for a in markata.articles if "config_overrides" in a.keys()]:
        raw_text = article.get("config_overrides", {}).get("head", {}).get("text", "")

        if isinstance(raw_text, list):
            article["config_overrides"]["head"]["text"] = "\n".join(
                flatten([t.values() for t in raw_text])
            )


@hook_impl
def render(markata: "Markata") -> None:
    if "post_template" in markata.config:
        template_file = markata.config["post_template"]
    else:
        template_file = Path(__file__).parent / "default_post_template.html"
    with open(template_file) as f:
        template = Template(f.read(), undefined=SilentUndefined)

    if "{{" in str(markata.config.get("head", {})):
        head_template = Template(
            str(markata.config.get("head", {})), undefined=SilentUndefined
        )
    else:
        head_template = None
        head = {}

    _full_config = copy.deepcopy(markata.config)

    for article in [a for a in markata.articles if hasattr(a, "html")]:

        if head_template:

            head = eval(
                head_template.render(
                    __version__=__version__,
                    config=_full_config,
                    **article,
                )
            )

        merged_config = {
            **_full_config,
            **{"head": head},
        }

        merged_config = always_merger.merge(
            merged_config,
            copy.deepcopy(
                article.get(
                    "config_overrides",
                    {},
                )
            ),
        )

        article.html = template.render(
            __version__=__version__,
            body=article.html,
            toc=markata.md.toc,  # type: ignore
            config=merged_config,
            **article.metadata,
        )



from pathlib import Path

import toml
from conftest import set_directory

from markata import Markata
from markata.plugins import post_template


def test_head_config_text_str(tmp_path):
    with set_directory(tmp_path):
        m = Markata()
        m.config["head"] = {}
        m.config["head"]["text"] = "here"
        post_template.configure(m)
        assert m.config["head"]["text"] == "here"


def test_head_config_text_dict(tmp_path):
    with set_directory(tmp_path):
        m = Markata()
        m.config["head"] = {}
        m.config["head"]["text"] = [{"value": "one"}, {"value": "two"}]
        post_template.configure(m)
        assert m.config["head"]["text"] == "one\ntwo"


def test_head_config_text_str_toml(tmp_path):
    with set_directory(tmp_path):
        Path("markata.toml").write_text(
            toml.dumps({"markata": {"head": {"text": "here"}}})
        )
        m = Markata()
        post_template.configure(m)
        assert m.config["head"]["text"] == "here"


def test_head_config_text_list_toml(tmp_path):
    with set_directory(tmp_path):
        Path("markata.toml").write_text(
            toml.dumps(
                {"markata": {"head": {"text": [{"value": "one"}, {"value": "two"}]}}}
            )
        )
        m = Markata()
        post_template.configure(m)
        assert m.config["head"]["text"] == "one\ntwo"



"""
Adds a service_worker to your site.  This will make it installable on mobile,
viewable offline, and potentially more responsive as the user goes between good
and bad connections.

## Configuration

Enable this plugin by adding it to your `markata.toml` hooks list.

``` toml
[markata]
hooks=[
  # your hooks
  "markata.plugins.service_worker",
]
```

If you have any content that you want to  precache, add it to the list of
precache.  You can use devtools, change your network to offline, and see what
files send 404's to the console.  These files likely need precache.

``` toml
[markata]
precache_urls = ['archive-styles.css', 'scroll.css', 'manifest.json']
```

## cache busting

Markata uses the checksum.dirhash of your output directory as the cache key.
This is likely to change and bust the cache on every build.

## pre-caching feeds

You can add and entire feed to your precache, this will automatically load
these posts into the cache anytime someone visits your site and their browser
installs the service worker.

Be nice to your users and don't try to install everything possible in their
cache, but maybe a few that they are most likely to click on.

``` toml
[markata.feeds.recent]
filter="date<today and date>today-timedelta(days=30) and published"
sort="slug"
precache=true
```

> note this assumes that the blog implements a published boolean in each posts
frontmatter.
"""
import copy
from pathlib import Path
from typing import TYPE_CHECKING

from checksumdir import dirhash
from jinja2 import Template

from markata import __version__
from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata

DEFAULT_PRECACHE_URLS = ["index.html", "./"]


@hook_impl(trylast=True)
def render(markata: "Markata") -> None:
    """
    sets precache_urls in markata.config to be used in
    `markata.plugins.service_worker.save`.
    """

    markata.config["precache_urls"] = markata.config.get("precache_urls", [])
    markata.config["precache_urls"].extend(DEFAULT_PRECACHE_URLS)

    for feed, config in markata.config.get("feeds").items():
        markata.config["precache_urls"].append(f"/{feed}/")

        if config.get("precache", False):
            for post in markata.map("post", **config):
                markata.config["precache_urls"].append(f'/{post.get("slug", "")}/')

    markata.config["precache_urls"] = list(set(markata.config["precache_urls"]))


@hook_impl(trylast=True)
def save(markata: "Markata") -> None:
    """
    Renders the service-worker.js file with your precache urls, and dirhash.
    """

    if "service_worker_template" in markata.config:
        template_file = markata.config["service_worker_template"]
    else:
        template_file = Path(__file__).parent / "default_service_worker_template.js"
    with open(template_file) as f:
        template = Template(f.read())

    output_dir = Path(markata.config.get("output_dir", "markout"))
    service_worker_file = output_dir / "service-worker.js"
    service_worker_js = template.render(
        __version__=__version__,
        config=copy.deepcopy(markata.config),
        output_dirhash=dirhash(output_dir),
    )

    service_worker_file.write_text(service_worker_js)
